﻿using System;
using System.Linq;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

using Apos.Shapes;

using Pong.Behaviours;

using Syntriax.Engine.Core;
using Syntriax.Engine.Core.Factory;
using Syntriax.Engine.Network;
using Syntriax.Engine.Physics2D;
using Syntriax.Engine.Systems.Tween;

namespace Pong;

public class GamePong : Game
{
    private readonly IUniverseObject platformSpecificUniverseObject = null!;
    private readonly GraphicsDeviceManager graphics = null!;
    private SpriteBatch spriteBatch = null!;
    private ShapeBatch shapeBatch = null!;

    private Universe universe = null!;
    private BehaviourCollector<IDisplayableSprite> displayableCollector = null!;
    private BehaviourCollector<IDisplayableShape> displayableShapeCollector = null!;
    private MonoGameCamera2DBehaviour cameraBehaviour = null!;
    private PongManagerBehaviour pongManager = null!;

    public GamePong(IUniverseObject platformSpecificUniverseObject)
    {
        this.platformSpecificUniverseObject = platformSpecificUniverseObject;
        graphics = new GraphicsDeviceManager(this)
        {
            PreferredBackBufferWidth = 1024,
            PreferredBackBufferHeight = 576,
            GraphicsProfile = GraphicsProfile.HiDef
        };

        Content.RootDirectory = "Content";
        IsMouseVisible = true;
    }

    protected override void Initialize()
    {
        // TODO: Add your initialization logic here
        universe = new();
        displayableCollector = new(universe);
        displayableShapeCollector = new(universe);

        universe.Initialize();

        base.Initialize();
    }

    protected override void LoadContent()
    {
        spriteBatch = new SpriteBatch(GraphicsDevice);
        shapeBatch = new ShapeBatch(GraphicsDevice, Content);
        SpriteFont spriteFont = Content.Load<SpriteFont>("UbuntuMono");

        universe.Register(platformSpecificUniverseObject);

        if (Environment.GetCommandLineArgs().FirstOrDefault(x => x.CompareTo("-server") == 0) is not null)
        {
            LiteNetLibServer server = universe.InstantiateUniverseObject<LiteNetLibServer>().SetUniverseObject("Server");
            UniverseObjectFactory.Instantiate<NetworkManager>().SetUniverseObject("NetworkManager", server.BehaviourController.UniverseObject);
            server.Start(8888, 2);
            Window.Title = $"Server";
        }
        else
        {
            LiteNetLibClient client = universe.InstantiateUniverseObject<LiteNetLibClient>().SetUniverseObject("Client");
            UniverseObjectFactory.Instantiate<NetworkManager>().SetUniverseObject("NetworkManager", client.BehaviourController.UniverseObject);
            client.Connect("localhost", 8888);
            Window.Title = $"Client";

            universe.InstantiateUniverseObject<DrawManager>().SetUniverseObject("Draw Manager");
        }

        universe.InstantiateUniverseObject<UpdateManager>().SetUniverseObject("Update Manager");
        universe.InstantiateUniverseObject<CoroutineManager>().SetUniverseObject("Coroutine Manager");
        universe.InstantiateUniverseObject<TweenManager>().SetUniverseObject("Tween Manager");
        universe.InstantiateUniverseObject<PhysicsEngine2D>().SetUniverseObject("Physics Engine 2D");

        ////////////////////////////////////////////////////////////////////////////////////

        cameraBehaviour = universe.InstantiateUniverseObject().SetUniverseObject("Camera")
            .BehaviourController.AddBehaviour<Transform2D>()
            .BehaviourController.AddBehaviour<CameraController>()
            .BehaviourController.AddBehaviour<MonoGameCamera2DBehaviour>(graphics);

        ////////////////////////////////////////////////////////////////////////////////////

        pongManager = universe.InstantiateUniverseObject().SetUniverseObject("Pong Game Manager")
            .BehaviourController.AddBehaviour<PongManagerBehaviour>(5);

        ////////////////////////////////////////////////////////////////////////////////////

        universe.InstantiateUniverseObject().SetUniverseObject("Ball")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(0, 0f), scale: new Vector2D(10f, 10f))
            .BehaviourController.AddBehaviour<CircleBehaviour>(new Circle(Vector2D.Zero, 1f))
            .BehaviourController.AddBehaviour<BallBehaviour>()
            .BehaviourController.AddBehaviour<RigidBody2D>();

        ////////////////////////////////////////////////////////////////////////////////////

        IUniverseObject leftPaddle = UniverseObjectFactory.Instantiate().SetUniverseObject("Left Paddle");
        leftPaddle.BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(-468f, 0f), scale: new Vector2D(15f, 60f))
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>();
        PaddleBehaviour leftPaddleBehaviour = BehaviourFactory.Instantiate<PaddleBehaviour>(Keys.W, Keys.S, 228f, -228f, 400f);
        leftPaddleBehaviour.Id = "lp";
        leftPaddle.BehaviourController.AddBehaviour(leftPaddleBehaviour);
        universe.Register(leftPaddle);

        IUniverseObject rightPaddle = UniverseObjectFactory.Instantiate().SetUniverseObject("Right Paddle");
        rightPaddle.BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(468f, 0f), scale: new Vector2D(15f, 60f))
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>();
        PaddleBehaviour rightPaddleBehaviour = BehaviourFactory.Instantiate<PaddleBehaviour>(Keys.Up, Keys.Down, 228f, -228f, 400f);
        rightPaddleBehaviour.Id = "rp";
        rightPaddle.BehaviourController.AddBehaviour(rightPaddleBehaviour);
        universe.Register(rightPaddle);

        ////////////////////////////////////////////////////////////////////////////////////

        universe.InstantiateUniverseObject().SetUniverseObject("Wall Top")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(0f, 308f), scale: new Vector2D(552f, 20f))
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;

        universe.InstantiateUniverseObject().SetUniverseObject("Wall Bottom")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(0f, -308f), scale: new Vector2D(552f, 20f))
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;

        universe.InstantiateUniverseObject().SetUniverseObject("Wall Right")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(532f, 0f), scale: new Vector2D(20f, 328f))
            .BehaviourController.AddBehaviour<WallScoreBehaviour>((Action)pongManager.ScoreToLeft)
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;

        universe.InstantiateUniverseObject().SetUniverseObject("Wall Left")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(-532f, 0f), scale: new Vector2D(20f, 328f))
            .BehaviourController.AddBehaviour<WallScoreBehaviour>((Action)pongManager.ScoreToRight)
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;

        ////////////////////////////////////////////////////////////////////////////////////

        universe.InstantiateUniverseObject().SetUniverseObject("Score Left")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(-250f, 250f), scale: Vector2D.One * .25f)
            .BehaviourController.AddBehaviour<TextScoreBehaviour>(true, spriteFont);

        universe.InstantiateUniverseObject().SetUniverseObject("Score Right")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(250f, 250f), scale: Vector2D.One * .25f)
            .BehaviourController.AddBehaviour<TextScoreBehaviour>(false, spriteFont);
    }

    protected override void Update(GameTime gameTime)
    {
        if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed || Keyboard.GetState().IsKeyDown(Keys.Escape))
            Exit();

        universe.Update(gameTime.ToUniverseTime());

        base.Update(gameTime);
    }

    protected override void Draw(GameTime gameTime)
    {
        GraphicsDevice.Clear(new Color() { R = 32, G = 32, B = 32 });

        // TODO: Add your drawing code here
        universe.Draw();

        spriteBatch.Begin(SpriteSortMode.Deferred, transformMatrix: cameraBehaviour.MatrixTransform);
        for (int i = displayableCollector.Behaviours.Count - 1; i >= 0; i--)
            displayableCollector.Behaviours[i].Draw(spriteBatch);
        spriteBatch.End();

        shapeBatch.Begin(cameraBehaviour.MatrixTransform);
        for (int i = displayableShapeCollector.Behaviours.Count - 1; i >= 0; i--)
            displayableShapeCollector.Behaviours[i].Draw(shapeBatch);
        shapeBatch.End();

        base.Draw(gameTime);
    }
}
